/*
 * Optimize an AA-Pattern using various methods.
 * Created by Abdalla Ahmed
 * 2014-12-01
 */

#include <CGAL/Exact_predicates_inexact_constructions_kernel.h>
#include <CGAL/Delaunay_triangulation_2.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <getopt.h>
#include <time.h>
#include <complex>


typedef CGAL::Exact_predicates_inexact_constructions_kernel K;
typedef K::Point_2 Point;
typedef K::Vector_2 Vector;
typedef std::vector<Point> Points;
typedef std::vector<Vector> Vectors;
typedef CGAL::Delaunay_triangulation_2<K>   DT;
typedef DT::Vertex_handle                   VH;
typedef DT::Face_handle                     FH;
typedef DT::Vertex_circulator               VC;
typedef DT::Face_circulator                 FC;

unsigned *shuffle(const unsigned N) {                                           // Return a randomly ordered list.
    unsigned *list = new unsigned[N];
    for (int i = 0; i < N; i++) list[i] = i;
    for (unsigned i = 0; i < N-1; i++) {
        unsigned r = i + rand() % (N-1 - i);
        std::swap(list[i], list[r]);
    }
    return list;
}

const double TWO_PI = 6.28318530717959;

struct Statistics {                                                             // This structure is adapted from psa code "https://code.google.com/p/psa/"
    float mindist;                                                              // Global minimum nearest neighbor distance
    float avgmindist;                                                           // Average minimum nearest neighbor distance
    float orientorder;                                                          // Bond orientation order
    float effnyquist;                                                           // Effective Nyquist frequency
    float oscillations;                                                         // Oscillations, please refer to Heck et al. for meaning of these measures
    float coverageRadius;                                                       // We added this to monitor the size of holes
    float N4, N5, N6, N7, N8;                                                   // We added this to track number of neighbors percentage
    Statistics() : mindist(0), avgmindist(0), orientorder(0),
    effnyquist(0), oscillations(0), coverageRadius(0),
    N4(0), N5(0), N6(0), N7(0), N8(0) {};
};

inline bool isInRect(const Point &p, const Point &BL, const Point &TR) {        // Check if a point is inside a given rectangle
    return BL.x() <= p.x() && BL.y() <= p.y() &&
    p.x() < TR.x() && p.y() < TR.y();
}

class AA {
private:
    int t, r, q, p, ONE;                                                        // Pattern parameters.
    int tt;                                                                     // t x t is the number of points in one period on an AA pattern
    double dhex;                                                                // Reference spacing of hexagonal packing.
    double davg = 0.87;                                                         // For use with spring relaxation
    Points map;                                                                 // A mapping from t-domain to pattern domain in a single period.
    struct TClampedGroup {                                                      // Points clamped to undergo similar shifts
        int xl, yb, xr, yt;                                                     // Block in t-domain. Bottom-left is included, top-right not.
        int n;                                                                  // Number of points in group.
        Vector v;                                                               // To aggregate shift vectors.
    } *g;                                                                       // A list of clamping groups
    int *gmap;                                                                  // Mapping back of t-domain to groups
    int ng = 0;                                                                 // Number of groups
    DT dt;                                                                      // Will maintain a Delaunay triangulation for relaxation and queries
    struct TSite {                                                              // This is to handle the Delaunay triangulation.
        Point p;
        VH vh[9];                                                               // Handles to up to 9 replicas to maintain toroidal domain
        inline double x() { return p.x(); };
        inline double y() { return p.y(); };
    };
    std::vector<TSite> v;                                                       // The final coordinates of points.
    inline double toroidallinearDist(double x1, double x2) const {              // 1D Nearest distance between replicas of two points
        double dx = x1 - x2;                                                    // Find distance in primary period
        if (dx > q) dx -= ONE;                                                  // If larger than half the period length another replica is certainly nearer
        if (dx < -q) dx += ONE;                                                 // Same, but opposite ordering of points
        return dx;
    };
    inline double toroidalSqDist(Point &p1, Point &p2) const {                  // 2D Nearest distance; square to avoid costly square root operation
        double dx = toroidallinearDist(p1.x(), p2.x());
        double dy = toroidallinearDist(p1.y(), p2.y());
        return dx * dx + dy * dy;
    };
    inline Point replica(Point p, int i) {                                      // Find one of the 9 replicas of a point
        i = (i+4) % 9;                                                          // We make the middle replica at index 0
        double x = p.x() + (i%3 - 1) * ONE;                                     // Add -ONE, 0, or ONE to x
        double y = p.y() + (i/3 - 1) * ONE;                                     // Same for y
        return Point(x, y);
    };
    Point marginBL, marginTR;                                                   // Points within these margins affect points in the pattern
    Point setSite(int index, Point p) {                                         // Set location of the indexed point (in t-domain) and insert it in triangulation
        v[index].p = p;                                                         // Save a handy copy of point coordinates
        for (int i = 0; i < 9; i++) {                                           // We loop through the 9 replicas,
            if (isInRect(replica(map[index], i), marginBL, marginTR))           // if the original (not the shifted) location of a replica is within margin
                v[index].vh[i] = dt.insert(replica(p, i));                      // insert replica in triangulation and keep handle to it
        }
        return p;
    };
    void moveSite(int index, Point p) {                                         // Adjust location of indexed point (in t-domain) and update in triangulation
        v[index].p = p;                                                         // Save a handy copy of updated point coordinates
        for (int i = 0; i < 9; i++) {
            if (isInRect(replica(map[index], i), marginBL, marginTR))           // Because points can move we use the original map for defining margins
                v[index].vh[i] = dt.move(v[index].vh[i], replica(p, i));
        }
    };
    inline void moveSite(int index, Vector& shift) {                            // A handy function to shift points relative to their current location
        moveSite(index, v[index].p + shift);
    };
    Points VoronoiVertices(int i) {                                             // Return the vertices of a Voronoi cell; needed by centroid()
        Points result;
        FC fc = dt.incident_faces(v[i].vh[0]), done(fc);
        do {
            result.push_back(dt.circumcenter(fc));
        } while (++fc != done);
        return result;
    };
    Points neighbors(int i) const {                                             // Return nearest points (in a Voronoi sense); needed by farthestPoint()
        Points result;
        VC vc = dt.incident_vertices(v[i].vh[0]), done(vc);
        do {
            result.push_back(vc->point());
        } while (++vc != done);
        return result;
    };
    Vectors relNeighbors(int i) {                                               // Neighbors in relative coordinates; needed by CCDT() and spring()
        Vectors result;
        VC vc = dt.incident_vertices(v[i].vh[0]), done(vc);
        do {
            result.push_back(vc->point() - v[i].p);
        } while (++vc != done);
        return result;
    };
    // Miscellaneous relaxation forces:
    typedef Vector (AA::*forceFunction) (int);                                  // A pointer to a member function for calculating forces & return shift vectors
    Vector centroid(int index) {                                                // See http://en.wikipedia.org/wiki/Centroid
        Points p = VoronoiVertices(index);                                      // A vector of ordered Voronoi cell vertices
        double a = 0, cx = 0, cy = 0;                                           // Cell area (actually twice the area) and centroid coordinates
        int n = p.size();                                                       // Number of vertices of Voronoi cell (= number of neighbors)
        double XProduct;                                                        // Cross product of vectors to adjacent vertices
        for (int j = 0; j < n; j++) {
            XProduct = p[j].x() * p[(j+1)%n].y() - p[j].y() * p[(j+1)%n].x();
            a += XProduct;                                                      // Accumulate areas
            cx += (p[j].x() + p[(j+1) % n].x()) * XProduct;
            cy += (p[j].y() + p[(j+1) % n].y()) * XProduct;
        }
        cx /= 3.0 * a;
        cy /= 3.0 * a;
        return Point(cx, cy) - v[index].p;                                      // Return shift from current position to centroid
    };
    Vector farthestPoint(int index) {                                           // A very localized version of farthest point optimization
        if (g[ gmap[index] ].n > 1)                                             // Since FP vector is not scalable, averaging won't work
            return spring(index) / 15;                                          // Use spring for grouped points, scale it very low so that FPO is dominant
        Points points = neighbors(index);                                       // We will search for farthest point in the convex hull of neighbors
        DT localDT(points.begin(), points.end());                               // We use an auxiliary Delaunay triangulation for this purpose
        Point result = v[index].p;                                              // A fall-back setting for result; this proved essential (but why??)
        double dmin = 0;
        DT::Finite_faces_iterator it;
        for (it = localDT.finite_faces_begin();
             it != localDT.finite_faces_end(); it++) {                          // Iterated through faces of the locally set up Delaunay triangulation
            Point c = localDT.circumcenter(it);                                 // Circumcenters of these faces are the candidate farthest points
            if (localDT.is_infinite(localDT.locate(c))) continue;               // A point in infinite face of local DT means it is outside its convex hull; skip
            double r = CGAL::squared_distance(c, it->vertex(0)->point());       // Distance to nearest points
            if (r > dmin) {                                                     // If larger than previous candidate
                result = c;                                                     // Make this the candidate
                dmin = r;                                                       // And its distance the candidate dmin
            }
        }
        return result - v[index].p;
    };
    Vector CCDT (int index) {                                                   // Capacity constrained Delaunay triangulation
        Vectors p = relNeighbors(index);                                        // List of nearest neighbors, using coordinates from concerned point
        int n = p.size();                                                       // Size of the list
        double a1, a2, b1, b2, c1, c2, det;                                     // Coefficients of linear system of minima
        double x1, y1, x2, y2;                                                  // Coordinates of successive neighbors
        a1 = a2 = b1 = b2 = c1 = c2 = 0;                                        // Clear coefficients
        for (int j = 0; j < n; j++) {                                           // Loop through neighbors
            x1 = p[j].x(); x2 = p[(j+1) % n].x();                               // Take a pair of consecutive neighbors
            y1 = p[j].y(); y2 = p[(j+1) % n].y();
            a1 += (y1 - y2) * (y1 - y2);                                        // Coefficient of x in partial derivative of area w.r.t x
            b1 += (x2 - x1) * (y1 - y2);                                        // Coefficient of y in partial derivative of area w.r.t x
            c1 += (y2 - y1) * (x1*y2 - x2*y1);                                  // (minus)Constant term in partial derivative of area w.r.t x
            a2 += (x2 - x1) * (y1 - y2);                                        // a2 = b1; coz it was the coefficients of xy before differentiation w.r.t x & y
            b2 += (x2 - x1) * (x2 - x1);                                        // Coefficient of y in partial derivative of area w.r.t y
            c2 += (x1 - x2) * (x1*y2 - x2*y1);                                  // (minus)Constant term in partial derivative of area w.r.t y
        }
        det = a1 * b2 - a2 * b1;                                                // Determinant of the linear system
        double x = (b2 * c1 - b1 * c2) / det;                                   // Solution for x
        double y = (a1 * c2 - a2 * c1) / det;                                   // Solution for y
        return Vector(x, y);                                                    // x and y are already relative to concerned point
    }
    Vector spring (int index) {                                                 // This is our home-made relaxation force: a special kind of spring force
        Vectors p = relNeighbors(index);                                        // List of nearest neighbors, using coordinates from concerned point
        int n = p.size();                                                       // Size of the list
        double springLength = davg * dhex;                                      // The target neighbor distance in absolute units
        Vector resultant = {0, 0};                                              // Resultant force
        for (int j = 0; j < n; j++) {                                           // Iterate through all neighbors; only close ones are actually involved
            double l = sqrt(p[j].squared_length());                             // Distance to current neighbor
            if (l < springLength) {                                             // If this neighbor is closer than target distance
                resultant = resultant + (1 - dhex/l) * p[j];                // The spring expands towards dhex, not its target length. Found experimentally
            }
        }
        return resultant;
    }
    inline Vector CCDTSpring (int index) {                                      // Jointly apply CCDT (reduces coverage radius) and spring (increases dmin)
        return (2 * CCDT(index) + spring(index)) / 3;                           // This ratio was found experimentally
    };
    inline Vector ctrdSpring (int index) {                                      // Jointly apply centroid and spring (increases dmin)
        return (2 * centroid(index) + spring(index)) / 3;                       // This ratio is experimental
    };
    inline Vector springOrCCDT (int index) {                                    // Apply spring if there are close neighbors, otherwise apply CCDT
        Vector s = spring(index);                                               // Try spring first
        if (s != Vector(0, 0)) return s;                                        // If it returns 0 (no neighbors closer than target Poisson dist radius)
        return CCDT(index);                                                     // switch to CCDT
    };
    inline Vector springOrCTRD (int index) {                                    // Apply spring if there are close neighbors, otherwise apply centroid
        Vector s = spring(index);                                               // Try spring first
        if (s != Vector(0, 0)) return s;                                        // If it returns 0 (no neighbors closer than target Poisson dist radius)
        return centroid(index);                                                 // switch to centroid
    };
    inline Point normalize(Point p) {                                           // Scale to unit & wrap around if points were shifted outside primary period
        double x = p.x() / ONE;
        double y = p.y() / ONE;
        while (x < 0) x += 1;
        while (y < 0) y += 1;
        while (x >= 1) x -= 1;
        while (y >= 1) y -= 1;
        return {x, y};
    };
    std::string outputPath = "";                                                // Path for output files; could be set to a folder in /tmp
public:
    enum ForceType {
        FT_Centroid = 0,
        FT_FarthestPoint = 1,
        FT_CCDT = 2,
        FT_SPRING = 3,
        FT_CCDTSpring = 4,
        FT_CTRD_SPRING = 5,
        FT_SPRING_OR_CCDT = 6,
        FT_SPRING_OR_CTRD = 7
    };
    AA(int t, int r) {                                                          // Constructor. We assume m = 1; so t and r fully specify the pattern
        this->t = t;
        this->r = r;
        q = t + r;
        p = q + t;
        ONE = 2*q;                                                              // This is the period of the pattern is source coordinates
        tt = t*t;                                                               // Number of points in one period; will be used very frequently
        dhex = ONE * sqrt(2/(sqrt(3) * tt));                                    // dhex in pattern-domain distances (source coordinates)
        double margin = 0.1;                                                    // Margin for toroidal domain. 10% should be quite enough
        marginBL = {-margin * ONE, -margin * ONE};                              // Bottom-left of primary period + margin
        marginTR = {(1+margin) * ONE, (1+margin) * ONE};                        // Top-right. In our convention BL is included, TR is excluded
        map.reserve(tt);                                                        // A mapping from t-domain to pattern-domain (in source coordinates)
        v.reserve(tt);                                                          // The final location including shifts
        int dX, dY;                                                             // Indexes in t-domain
        for (int y = 0; y < ONE; y++) {                                         // Transforming from t-domain to pattern is difficult; instead we
            for (int x = 0; x < ONE; x++) {                                     // loop through one period in pattern-domain and map points to t-domain
                dX = (p*x + q*y) % ONE;                                         // For simplicity we set the t-domain at bottom-left;
                dY = (p*y + q*x) % ONE;                                         // see jgt paper by Ahmed.
                if (dX < t && dY < t) {                                         // if x & y satisfy conditions
                    Point p = {(double)x, (double)y};
                    map[t*dY + dX] = p;                                         // Insert point in the appropriate index in the t-domain map
                    setSite(t*dY + dX, p);                                      // and insert the point in the Delaunay triangulation
                }
            }
        }
        // Setup groups; we simply put each point in its own group
        gmap = new int [tt];                                                    // A map to show the group of each point
        g = new TClampedGroup [tt];                                             // A list of groups
        ng = tt;                                                                // Initialize number of groups to number of points
        for (int i = 0; i < tt; i++) {                                          // Loop through all points
            gmap[i] = i;                                                        // Put point #i in group #i.
            g[i] = {i % t, i / t, i%t + 1, i / t + 1, 1};                       // Initialize group #i to point #i and its point count to 1.
        }
    };
    void setOutputPath(std::string path) { outputPath = path; };                // Let the user choose output folder. May be /tmp/??
    void setSpringLength(double d) { davg = d; };                               // Set target NND for spring().
    void loadGroupMap(const char *fileName) {                                   // Read grouping map from pgm file. It should simply be a checkered map.
        FILE *pgmfile = fopen(fileName, "r");
        if (!pgmfile) {
            fprintf(stderr, "Failed to open file %s\n", fileName);
            exit(1);
        }
        int w, h, grades;                                                       // Width, height, and grades of pgm file
        fscanf(pgmfile, "P5 %d %d %d ", &w, &h, &grades);                       // We assume no comments in pgm file
        if (w != t || h != t) {
            fprintf(stderr, "map must be t x t\n");
            exit(2);
        }
        unsigned char *bmp = new unsigned char[tt];                             // We will read the map in a bitmap
        unsigned char c;
        for (int i = 0; i < tt ; i++) {
            fscanf(pgmfile, " %c", &bmp[i]);
            if (feof(pgmfile)) {
                fprintf(stderr, "Error: Incomplete map file %s\n", fileName);
                exit(1);
            }
        }
        ng = 0;                                                                 // Reset number of groups; will be incremented subsequently
        for (int i = 0; i < tt; i++) gmap[i] = -1;                              // Invalidate all group assignments
        int xl, yb, xr, yt;
        for (int i = 0; i < tt; i++) {
            if (gmap[i] != -1) continue;                                        // Already assigned to a group? skip
            xl = i % t;                                                         // Start a new block
            yb = i / t;
            for (xr = xl+1; xr < t && bmp[yb*t+xr] == bmp[i]; xr++) ;           // Scan for invariance along row
            for (yt = yb+1; yt < t && bmp[yt*t + xl] == bmp[i]; yt++) ;         // ~ along column
            for (int y = yb; y < yt; y++) {                                     // Now mark this group in gmap
                for (int x = xl; x < xr; x++) {
                    gmap[y*t + x] = ng;
                }
            }
            g[ng++] = {xl, yb, xr, yt, (xr-xl)*(yt-yb)};                        // And set its parameters in group list
        }
        fprintf(stderr, "Read %d regions\n from %s\n", ng, fileName);           // For debugging
        delete[] bmp;
    };
    double printVectors(const char* fileName) {                                 // Print shift vectors of groups; for use by rollout program
        const char *fullFileName = (outputPath + fileName).c_str();             // Include path and/or prefixes
        FILE *file = fopen(fullFileName, "w");
        if (!file) {
            fprintf(stderr, "Failed to open %s\n", fullFileName);
            exit(1);
        }
        double maxShift = 0;                                                    // Maximum distance traveled by points after relaxation
        for (int i = 0; i < ng; i++) {                                          // Iterate through groups
            int j = g[i].yb * t + g[i].xl;                                      // Index of the first point in the group block; to represent the group.
            Vector d = v[j].p - map[j];                                         // Shift made by this point; in source domain coordinates
            fprintf(file, "%lf %lf\n", d.x(), d.y());                           // Print it out
            if (maxShift < d.squared_length())
                maxShift = d.squared_length();
        }
        fclose(file);
        return sqrt(maxShift);
    }
    void loadVectors(const char *fileName) {                                    // Load shift vectors saved by printVectors; useful for resuming optimization
        FILE *vectorsFile = fopen(fileName, "r");
        if (!vectorsFile) {
            fprintf(stderr, "Failed to open %s\n", fileName);
            exit(1);
        }
        for (int i = 0; i < ng; i++) {                                          // Loop through all groups (in the same order as for saving)
            double dx, dy;
            fscanf(vectorsFile, " %lf %lf", &dx, &dy);                          // Read a shift vector entry
            if (!feof(vectorsFile)) {                                           // If successful read
                g[i].v = {dx, dy};                                              // Assign it to group vector
            } else {                                                            // End of file reached?
                fprintf(stderr, "Failed to read all vectors from %s\n",
                        fileName);
                exit(1);
            }
        }
        dt.clear();                                                             // Clear the Delaunay triangulation to rebuild from scratch
        for (int i = 0; i < tt; i++) {                                          // Loop through all points and insert them appropriately in DT
            setSite(i, map[i] + g[ gmap[i] ].v);
        }
    };
    void plotEPS(std::string fileName, double radius = 0.5) {                   // Plot EPS file of (shifted) points
        const char *fullFileName = (outputPath + fileName).c_str();
        FILE *epsfile = fopen(fullFileName, "w");
        if (!epsfile) {
            fprintf(stderr, "Failed to open %s\n", fullFileName);
            exit(1);
        }
        fprintf(epsfile, "%%!PS-Adobe-3.0 EPSF-3.0\n");
        fprintf(epsfile, "%%%%BoundingBox: 0 0 %d %d\n", ONE, ONE);             // Plot a single period
        fprintf(epsfile, "%d dup scale\n", ONE);                                // We scale down to unit torus to follow the convention
        fprintf(epsfile, "/r %f def\n", radius / ONE);
        fprintf(epsfile, "/p {r 0 360 arc fill} def\n");                        // A typical PS function for point plotting
        for (int i = 0; i < tt; i++) {                                          // Loop through all points
            Point p = normalize(v[i].p);                                        // Normalize and wrap back to unit torus
            fprintf(epsfile, "%f %f p\n", p.x(), p.y());                        // Print out
        }
        fprintf(epsfile, "showpage\n");                                         // Some viewer (e.g. new version of evince) may need this to display properly
        fclose(epsfile);                                                        // We need to close the file explicitly coz we are generating so many files
    };
    void printText(std::string fileName) {                                      // Generate a text printout
        const char *fullFileName = (outputPath + fileName).c_str();
        FILE *file = fopen(fullFileName, "w");
        if (!file) {
            fprintf(stderr, "Failed to open %s\n", fullFileName);
            exit(1);
        }
        fprintf(file, "%d\n", tt);                                              // Print #point is first line; a common convention
        for (int i = 0; i < tt; i++) {                                          // Loop through all points
            Point p = normalize(v[i].p);                                        // Normalize and wrap back to unit torus
            fprintf(file, "%f %f\n", p.x(), p.y());
        }
        fclose(file);
    };
    void shake(double jitter) {                                                 // Apply random jitter to points; needed to break the structure of the pattern
        for (int i = 0; i < ng; i++) {                                          // Loop through groups
            double dx = (drand48() * jitter - jitter/2.0);                      // Generate a random x offset
            double dy = (drand48() * jitter - jitter/2.0);                      // and a random y offset
            g[i].v = {dx, dy};                                                  // Save random vector to group, to be applied subsequently to points
        }
        for (int i = 0; i < tt; i++) {                                          // Loop through all points
            moveSite(i, g[ gmap[i] ].v);                                        // Apply to each point the shift vector of the respective group
        }
    }
    double relax(ForceType ft = FT_Centroid,                                    // Relax points using the chosen force
                 double scale = 1.0,                                            // Parameter passed to scale (attenuate) forces; 0.05 gave us good results
                 bool serial = false) {                                         // Apply forces one point (group) at a time rather than simultaneously?
        forceFunction forces[] = {                                              // List of defined force functions
            &AA::centroid,
            &AA::farthestPoint,
            &AA::CCDT,
            &AA::spring,
            &AA::CCDTSpring,
            &AA::ctrdSpring,
            &AA::springOrCCDT,
            &AA::springOrCTRD
        };
        forceFunction f = forces[(int)ft];                                      // Choose user-specified force function
        if (ft == FT_FarthestPoint) scale = 1.0;                                // FPO isn't scalable: the shift has to be made in full; period.
        double maxShift = 0;                                                    // To keep track of maximum shift in this iteration
        unsigned *seq = shuffle(ng);                                            // Use a random sequence to iterate through groups; for serial application
        for (int i = 0; i < ng; i++) {                                          // Iterate through group indexes
            int j = seq[i];                                                     // Translate index according to the random sequence
            g[j].v = {0, 0};                                                    // Clear the group vector to accumulate forces
            for (int y = g[j].yb; y < g[j].yt; y++) {                           // Iterate through all points in the group
                for (int x = g[j].xl; x < g[j].xr; x++) {
                    int index = y*t + x;
                    g[j].v = g[j].v + (this->*f)(index);                        // Find force on each point and add it to the resultant force on the group
                }
            }
            g[j].v = scale * g[j].v / g[j].n;                                   // Average and scale force on the group
            if (serial) {                                                       // If serial mode apply shifts immediately
                for (int y = g[j].yb; y < g[j].yt; y++) {                       // Loop again through all points in the group
                    for (int x = g[j].xl; x < g[j].xr; x++) {
                        int index = y*t + x;
                        moveSite(index, g[j].v);                                // Apply shift to each point and update the DT
                    }
                }
            }
            if (maxShift < g[j].v.squared_length())                             // This works irrespective of the mode (serial/simultaneous)
                maxShift = g[j].v.squared_length();
        }
        if (!serial) {                                                          // If points should be updated simultaneous
            for (int i = 0; i < tt; i++) {                                      // Loop through all points
                moveSite(i, g[ gmap[i] ].v);                                    // Apply to each point the force on its group and update the triangulation
            }
        }
        delete[] seq;
        return sqrt(maxShift);                                                  // Return maximum shift which can be used to monitor convergence
    };
    Statistics GetStatistics() const {                                          // Return statistics useful to monitor progress of optimization
        Statistics stats;                                                       // This function is adapted from PSA code.
        stats.mindist = ONE;                                                    // The minimum distance can't go above ONE :)
        stats.avgmindist = 0;                                                   // Reset this to 0 to accumulate for averaging
        stats.orientorder = 0;                                                  // ~
        std::complex<double> acc = 0;
        unsigned long nacc = 0;
        for (int i = 0; i < tt; i++) {
            float localmd = ONE;
            std::complex<double> localacc = 0;
            VC vc = dt.incident_vertices(v[i].vh[0]),
                done(vc), next;
            do {
                next = vc; ++next;
                const Point &v1 = vc->point();
                const Point &v2 = next->point();
                // Local mindist
                double dist = CGAL::squared_distance(v[i].p, v1);
                localmd = std::min(localmd, (float) dist);
                // Orientational order
                std::complex<double> c1(v1.x(), v1.y());
                std::complex<double> c2(v2.x(), v2.y());
                localacc += std::polar(1.0, 6.0 * arg(c1 - c2));                // This is how it's done in PSA (divide only at end); but I feel it needs review
                ++nacc;
            } while (++vc != done);
            stats.mindist = std::min(stats.mindist, localmd);
            stats.avgmindist += sqrtf(localmd);
            acc += abs(localacc);
        }
        // -------- Coverage Radius
        double Rc = 0;
        Point BL(0, 0), TR(ONE, ONE);                                           // We will consider only faces in one period
        DT::Finite_faces_iterator it;
        for (it = dt.finite_faces_begin(); it != dt.finite_faces_end(); it++) { // Iterate through all finite faces in triangulation
            if (isInRect(it->vertex(0)->point(), BL, TR)) {                     // It suffices to have one vertex within main period to consider the face
                Point c = dt.circumcenter(it);                                  // Farthest point in current face
                double r = CGAL::squared_distance(c, it->vertex(0)->point());   // Farthest distance in current face
                if (r > Rc) { Rc = r; }                                         // If larger than current farthest distance update the latter
            }
        }
        // -------- Voronoi cell N-Gon types:                                   // Before orientation order this used to be a measure; see paper by Balzer et al.
        int histogram[10] = {0,0,0,0,0,0,0,0,0,0};                              // A simple histogram, where index 0 is interpreted as 3
        for (int i = 0; i < tt; i++) {                                          // Iterate through all points
            VC vc = dt.incident_vertices(v[i].vh[0]), done(vc);                 // We did not call neighbors to save the overhead of allocating Points
            int n = 0; do { n++; } while (++vc != done);                        // Count number of neighbors
            ++histogram[n - 3];                                                 // Increment the respective histogram bin; triangle is the smallest possible cell
        }
        // -------- Aggregate statistics:
        stats.mindist = sqrtf(stats.mindist) / dhex;                            // Nearest Neighbor Distance (NND) should
        stats.avgmindist /= tt * dhex;                                          // be relative to dhex, the NND of hexagonal lattice
        stats.orientorder = abs(acc) / nacc;                                    // I don't quite get why we should average this once rather than average averages
        stats.coverageRadius = sqrtf(Rc) / dhex;                                // Size of largest hole (cf dmin); there should be an average too (cf davg)!
        stats.N4 = (float)histogram[1] / tt;                                    // Ratios of various Voronoi cell types
        stats.N5 = (float)histogram[2] / tt;                                    // 3 and 9+ are quite unlikely after relaxation
        stats.N6 = (float)histogram[3] / tt;
        stats.N7 = (float)histogram[4] / tt;
        stats.N8 = (float)histogram[5] / tt;
        return stats;
    };
};

const char *USAGE_MESSAGE = "Usage: %s [options] <t> <r>\n"
    "Options:\n"
    "-n <number of iterations>\n"
    "-p <path/prefix of output files>\n"
    "-g <path to grouping map file>\n"
    "-l <path to vectors table to load>\n"
    "-j <size of pre-optimization jitter (in grid scale)>\n"
    "-s <scale applied to forces>\n"
    "-d <target dmin for spring>\n"
    "-e order eps plots\n"
    "-t order text printouts\n"
    "-S orders serial application of forces\n"
;

int main(int argc,char **argv) {
    srand(time(NULL)); srand48(time(NULL));                                     // Random seeds to random number generators (int and real)
    int opt;                                                                    // For use by getopt, the command line options utility
    int iterations = 1;                                                         // Number of iterations to apply
    char *fileNamePrefix = (char *)"";                                          // Current working directory is default output directory
    bool printText = false;                                                     // Whether to make text printouts
    char *groupingFileName = NULL;                                              // Path to pgm grouping map
    double scale = 1.0;                                                         // Scale applied to forces. From experience We prefer damping the forces (~5%).
    double jitter = 0;                                                          // Size of jitter applied prior to optimization; a value around dhex is good
    char *vectorsFileName = NULL;                                               // Path to file for loading a shift-vectors table
    bool serial = false;                                                        // Whether to apply shifts serially (FPO style), we found it to improve isotropy
    bool printEPS = false;                                                      // Whether to make EPS plots
    double davg = 0.87;                                                         // Target dmin for use by our spring relaxation
    int ft = (int)AA::FT_Centroid;                                              // The force to use
    while ((opt = getopt(argc, argv, "n:p:g:l:j:k:s:d:etS")) != -1) {           // Modify default settings with command line options
        switch (opt) {
            case 'n': iterations = atoi(optarg); break;
            case 'p': fileNamePrefix = optarg; break;
            case 'g': groupingFileName = optarg; break;
            case 'l': vectorsFileName = optarg; break;
            case 'j': jitter = atof(optarg); break;
            case 'k': ft = atoi(optarg); break;
            case 's': scale = atof(optarg); break;
            case 'd': davg = atof(optarg); break;
            case 'e': printEPS = true; break;
            case 't': printText = true; break;
            case 'S': serial = true; break;
            default: fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
        }
    }
    if (optind > argc - 2) {
        fprintf(stderr, USAGE_MESSAGE, argv[0]); exit(1);
    }
    int t = atoi(argv[optind]);
    int r = atoi(argv[optind+1]);
    // Get grouing info, if any:
    AA aa(t, r);                                                                // Create the pattern object
    aa.setOutputPath(fileNamePrefix);                                           // Set path to output files (typically a temporary folder)
    if (groupingFileName != NULL) aa.loadGroupMap(groupingFileName);            // If a neighborhood map is supplied load and apply it
    if (vectorsFileName != NULL) aa.loadVectors(vectorsFileName);               // If a vectors table is supplied load and apply it
    if (printEPS) { aa.plotEPS("original.eps"); }                               // If EPS plots are ordered, make a plot prior to optimization
    if (jitter != 0) aa.shake(jitter);                                          // If jittering ordered apply it. Should be ordered unless a vector table loaded
    aa.setSpringLength(davg);                                                   // Set target dmin for spring-based relaxations
    char fName[200];                                                            // A buffer to compile output file names
    for (int i = 0; i < iterations; i++) {                                      // The main optimization loop
        double iterationMax = aa.relax((AA::ForceType)ft, scale, serial);       // Iterate and read back the maximum shift, useful to monitor convergence
        if (printEPS) { sprintf(fName, "%03d.eps", i); aa.plotEPS(fName); }     // Make EPS plot is ordered
        if (printText) { sprintf(fName, "%03d.txt", i); aa.printText(fName); }  // Make text printout if ordered
        sprintf(fName, "%03d.dat", i); double max = aa.printVectors(fName);     // Print shift vectors. This is the main output, all others can be derived from it
        Statistics stats = aa.GetStatistics();                                  // Read statistics after this iteration
        fprintf(stderr, "%4d- max-step = %9.6f, "                               // Print stats. It could be convenient to pipeline stderr to file then "tail -f"
            "max-shift = %9.6f, "
            "N4-N8%% = {%5.1f,%5.1f,%5.1f,%5.1f,%5.1f}, "
            "Rc = %.3f, G-MD = %.3f, A-MD = %.3f, BOO = %.3f\n",
            i, iterationMax, max, 100 * stats.N4, 100 * stats.N5,
            100 * stats.N6, 100 * stats.N7, 100 * stats.N8,
            stats.coverageRadius,
            stats.mindist, stats.avgmindist,
            stats.orientorder
        );
    }
}
